""" LocationsReport conveniences """
import http
import logging
from abc import ABC, abstractmethod
from enum import Enum
from pathlib import Path
from typing import NamedTuple, Optional

# pylint: disable=E0239, E0401, E0402, E1136, R0201, R0902, R0903, R0913, W0613
import requests
from marshmallow import Schema, fields, post_load
from marshmallow.decorators import validates_schema
from pycapo import CapoConfig

from .exceptions import FetchError, LocatorServiceException
from .interfaces import LocatedFile, LocationReport, Locator
from .ngas import NgasConnection

logger = logging.getLogger("productfetcher.locations")
logger.setLevel(logging.INFO)


class Location(Enum):
    """
    Where the files live
    """

    DSOC = "DSOC"
    NAASC = "NAASC"

    @property
    def lustre_prefix(self) -> str:
        """
        The expected Lustre path prefix for this location
        """
        return "/lustre/aoc" if self == self.DSOC else "/lustre/naasc"

    @staticmethod
    def in_location(path: Path) -> Optional["Location"]:
        """
        Given a path, return the location it appears to be in, or nothing,
        if you're out in space somewhere.

        :param path:  the path to test
        :return:  a Location, optionally
        """

        # look through the locations until we find one that seems to match
        for location in Location:
            if str(path.absolute()).startswith(location.lustre_prefix):
                return location

        return None


class Cluster(Enum):
    """
    Which cluster the files are on
    """

    DSOC = "DSOC"
    NAASC = "NAASC"


class NgasServer(NamedTuple):
    """
    Encapsulates the "server" section of a "files" item in a location report
    """

    server: str
    location: Location
    cluster: Cluster

    def can_direct_copy_to(self, dest: Path) -> bool:
        """
        True if we can direct copy to this destination.

        In principle, direct copy is possible if the file's NGAS server and the destination
        are in the same datacenter, but it also depends on the correct version of NGAS and
        the NGAS direct copy plugin, which is not supported by the NAASC cluster at the moment.
        So there is a test in here which will need to be removed eventually, to ensure that
        the DSOC cluster is accounted for.

        :param dest:  the destination to write to
        :return:      true if direct copy is possible
        """

        return self.cluster == Cluster.DSOC and self.location == Location.in_location(dest)

    def connect(self) -> NgasConnection:
        """
        Connect to this NGAS server and return an NgasConnection we can use to send requests to.

        :return: a connection to NGAS
        """
        return NgasConnection(self.server)

    def __str__(self):
        return f"NGAS {self.server} at {self.location}"


class NgasFile(LocatedFile):
    """Encapsulates all the information about an NGAS file in a locations report"""

    ngas_file_id: str
    subdirectory: str
    relative_path: str
    checksum: int
    checksum_type: str
    version: int
    size: int
    server: NgasServer

    def __init__(
        self,
        ngas_file_id,
        subdirectory,
        relative_path,
        checksum,
        checksum_type,
        version,
        size,
        server,
    ):
        super().__init__()
        self.ngas_file_id = ngas_file_id
        self.subdirectory = subdirectory
        self.relative_path = relative_path
        self.checksum = checksum
        self.checksum_type = checksum_type
        self.version = version
        self.size = size
        self.server = server

    def make_fetcher(self, factory: "FetcherFactory") -> "FileFetcher":
        return factory.fetch_ngas(self)

    def shares_resources_with(self, other: "LocatedFile") -> bool:
        return hasattr(other, "server") and self.server == other.server

    def can_direct_copy(self):
        """
        True if this file can be direct copied from NGAS.
        """
        return self.server.can_direct_copy_to(self.destination)

    def direct_copy(self):
        """
        Direct copy this file to its destination

        :return: None, if successful
        """
        with self.server.connect() as connection:
            connection.direct_copy(self.ngas_file_id, self.version, self.destination)

    def stream(self):
        """
        Stream this file into its destination.

        :return: None, if successful
        """
        with self.server.connect() as connection:
            connection.stream(self.ngas_file_id, self.version, self.destination)

    def __str__(self):
        return f"NGAS {self.ngas_file_id} from {self.server} -> {self.subdirectory}/{self.relative_path}"

    def __eq__(self, other):
        return (
            self.ngas_file_id == other.ngas_file_id
            and self.subdirectory == other.subdirectory
            and self.relative_path == other.relative_path
            and self.checksum == other.checksum
            and self.checksum_type == other.checksum_type
            and self.version == other.version
            and self.size == other.size
            and self.server == other.server
        )


class OracleXml(LocatedFile):
    """Represents the metadata of an ALMA SDM stored as XML in the ALMA DB"""

    archive_uid: str
    table: str
    subdirectory: Path
    relative_path: str
    size: int

    def __init__(self, archive_uid, table, subdirectory, relative_path, size):
        super().__init__()
        self.archive_uid = archive_uid
        self.table = table
        self.subdirectory = subdirectory
        self.relative_path = relative_path
        self.size = size

    def make_fetcher(self, factory: "FetcherFactory") -> "FileFetcher":
        return factory.fetch_oracle_xml(self)

    def shares_resources_with(self, other: "LocatedFile") -> bool:
        # all OracleXml files use the same resource
        return hasattr(other, "table")

    def __str__(self):
        return f"ALMA {self.archive_uid} @ {self.table} -> {self.subdirectory}/{self.relative_path}"


class FileLocator(Locator):
    """Loads a locations report from a .json report file"""

    def __init__(self, file: Path):
        self.file = file

    def locate(self) -> LocationReport:
        return LocationReportSchema().loads(self.file.read_text())


class ServiceLocator(Locator):
    """Acquires a locations report from the archiveService"""

    def __init__(self, science_product_locator: str):
        self.spl = science_product_locator

    def locate(self) -> LocationReport:
        settings = CapoConfig().settings("edu.nrao.workspaces.ProductFetcherSettings")

        # this is needed to prevent SSL errors when tests are run
        # inside a Docker container
        requests.packages.urllib3.util.ssl_.DEFAULT_CIPHERS += ":HIGH:!DH:!aNULL"
        requests.Session().mount(
            settings.locatorServiceUrlPrefix, adapter=requests.adapters.HTTPAdapter()
        )

        try:
            response = requests.get(settings.locatorServiceUrlPrefix, params={"locator": self.spl})
        except requests.exceptions.Timeout as exc_to:
            raise LocatorServiceException("Locator service timed out") from exc_to
        except requests.exceptions.TooManyRedirects as exc_re:
            raise FetchError("Too many redirects trying to reach locator service") from exc_re
        except requests.exceptions.RequestException as ex:
            raise FetchError(ex.response) from ex

        if response.status_code == http.HTTPStatus.OK:
            return LocationReportSchema().load(response.json())
        if response.status_code == http.HTTPStatus.NOT_FOUND:
            raise FetchError(f'locator "{self.spl}" not found')
        raise LocatorServiceException(f"Locator service returned {response.status_code}")


class NgasServerSchema(Schema):
    """marshmallow schema to interpret "server" section of a "files" item"""

    server = fields.Str()
    location = fields.Str()
    cluster = fields.Str()

    @post_load
    def make_ngas_server(self, data, **kwargs):
        return NgasServer(data["server"], Location[data["location"]], Cluster[data["cluster"]])


class NgasFileSchema(Schema):
    """One of the items in a location report's "files" list"""

    ngas_file_id = fields.Str()
    subdirectory = fields.Str(allow_none=True)
    relative_path = fields.Str()
    checksum = fields.Integer()
    checksum_type = fields.Str()
    version = fields.Integer()
    size = fields.Integer()
    server = fields.Nested(NgasServerSchema())
    table = fields.Str()
    archive_uid = fields.Str()

    @validates_schema
    def validate_subdirectory(self, data, **kwargs):
        # Fill in a subdirectory value that is None with the default
        if data["subdirectory"] is None:
            data["subdirectory"] = "retrieved-products"

    @post_load
    def make_filespec(self, data, **kwargs):
        if "table" in data:
            return OracleXml(**data)
        return NgasFile(**data)


class LocationReportSchema(Schema):
    """Encapsulates an entire locations report"""

    files = fields.List(fields.Nested(NgasFileSchema()))
    aggregate_size = fields.Integer()

    @post_load
    def make_report(self, data, **kwargs):
        return LocationReport(**data)
