""" Unit tests for Capability Service """

from unittest.mock import MagicMock, patch

import pytest

from workspaces.capability.schema import Capability, CapabilityExecution
from workspaces.capability.services.capability_info import CapabilityInfo
from workspaces.capability.services.capability_service import CapabilityService

# pylint: disable=C0301, E0401, R0201

pytest_plugins = ["testing.utils.conftest"]


@pytest.mark.usefixtures("mock_capability_service")
class TestCapabilityService:
    """Tests for CapabilityService methods"""

    def test_pause_capability(self, mock_capability_service: CapabilityService):
        # Sub lookup_capability with mock
        orig_lookup = mock_capability_service.capability_info.lookup_capability
        fake_capability = Capability(name="test", steps="test", max_jobs=2, paused=False)
        mock_capability_service.capability_info.lookup_capability = MagicMock(return_value=fake_capability)

        mock_capability_service.pause_capability("test")
        mock_capability_service.capability_info.lookup_capability.assert_called_with("test")
        assert fake_capability.paused is True
        mock_capability_service.capability_info.save_entity.assert_called_with(fake_capability)

        # Put original method back in place
        mock_capability_service.capability_info.lookup_capability = orig_lookup

    def test_unpause_capability(self, mock_capability_service: CapabilityService):
        # Sub lookup_capability with mock
        orig_lookup = mock_capability_service.capability_info.lookup_capability
        fake_capability = Capability(name="test", steps="test", max_jobs=2, paused=True)
        mock_capability_service.capability_info.lookup_capability = MagicMock(return_value=fake_capability)

        orig_startup = mock_capability_service.execution_manager.queue_manager.startup
        mock_capability_service.execution_manager.queue_manager.startup = MagicMock()

        mock_capability_service.unpause_capability("test")
        mock_capability_service.capability_info.lookup_capability.assert_called_with("test")
        assert fake_capability.paused is False
        mock_capability_service.capability_info.save_entity.assert_called_with(fake_capability)
        mock_capability_service.execution_manager.queue_manager.startup.assert_called()

        # Put original methods back in place
        mock_capability_service.capability_info.lookup_capability = orig_lookup
        mock_capability_service.execution_manager.queue_manager.startup = orig_startup

    @pytest.mark.skip("Broken due to queue/messenger rework. Does work in local `make test`")
    def test_on_ingestion_complete(
        self,
        mock_capability_service: CapabilityService,
        mock_capability_info: CapabilityInfo,
        mock_capability_execution: CapabilityExecution,
    ):
        fake_subject = {
            "type": "WorkflowRequest",
            "workflow_request_id": -1,
            "execution_wf_id": "-1",
        }

        fake_ingestion_complete_msg = {
            "subject": fake_subject,
            "service": "workflow",
            "routing_key": "workflow",
            "type": "ingestion-complete",
        }
        save_entity_old_call_count = mock_capability_info.save_entity.call_count
        mock_capability_service.on_ingestion_complete(**fake_ingestion_complete_msg)
        assert mock_capability_info.save_entity.call_count == save_entity_old_call_count + 1
        (request,) = mock_capability_info.save_entity.call_args.args
        assert request.ingested is True

    @pytest.mark.skip("As above, broken due to queue/messenger rework? Succeeds locally; fails on CI")
    def test_on_carta_ready(
        self,
        mock_capability_service: CapabilityService,
        mock_capability_info: CapabilityInfo,
        mock_capability_execution: CapabilityExecution,
    ):
        """
        Are we catching the "carta-ready" message and saving the metadata
         to the capability request version?

        :param mock_capability_service: stand-in for capability service
        :param mock_capability_info: stand-in for capability info
        :param mock_capability_execution: stand-in for capability execution
        :return:
        """

        wf_request_id = -2
        carta_url = "decartes_image_carta_url"
        fake_carta_ready_msg = {
            "service": "capability",
            "routing_key": "capability",
            "carta_url": carta_url,
            "subject": {"workflow_request_id": wf_request_id, "argument": {"parent_wf_request_id": -1}},
            "type": "carta-ready",
        }

        save_entity_old_call_count = mock_capability_info.save_entity.call_count

        with patch(
            "workspaces.capability.services.capability_info.CapabilityInfo.lookup_execution_by_workflow_request_id",
            return_value=mock_capability_execution,
        ):
            mock_capability_service.on_carta_ready(**fake_carta_ready_msg)
            assert mock_capability_info.save_entity.call_count == save_entity_old_call_count + 1

            (request_version,) = mock_capability_info.save_entity.call_args.args
            assert request_version.version_number > 0
            assert request_version.workflow_metadata["carta_url"] == carta_url
