import logging
from typing import Dict, List

import transaction
from messaging.messenger import MessageSender
from messaging.router import Router, on_message

from workspaces.capability.enums import CapabilityRequestState
from workspaces.capability.helpers import Parameter
from workspaces.capability.message_architect import CapabilityMessageArchitect
from workspaces.capability.schema import CapabilityRequest
from workspaces.capability.schema_interfaces import (
    CapabilityExecutionIF,
    CapabilityVersionIF,
)
from workspaces.capability.services.execution_manager import ExecutionManager
from workspaces.capability.services.interfaces import (
    CapabilityInfoIF,
    CapabilityServiceIF,
)
from workspaces.notification.services.interfaces import NotificationServiceIF
from workspaces.workflow.services.interfaces import WorkflowServiceIF

# pylint: disable=E0401, R0903, W1203


logger = logging.getLogger(__name__)


class CapabilityService(CapabilityServiceIF):
    """
    The capability service: clients access this to request capability runs
    """

    def __init__(
        self,
        capability_info: CapabilityInfoIF,
        workflow_service: WorkflowServiceIF,
        notification_service: NotificationServiceIF,
    ):
        self.messenger = MessageSender("capability")
        self.router = Router("capability")
        self.router.register(self)
        self.execution_manager = ExecutionManager(capability_info, workflow_service, self.router, self.messenger)
        self.capability_info = capability_info
        self.notify = notification_service

    def create_new_version(self, capability_request_id: int, parameters: List[Parameter] = None) -> CapabilityVersionIF:
        logger.info(f"Creating new version of Capability Request #{capability_request_id}")
        return self.capability_info.create_capability_version(capability_request_id, parameters)

    def pause_capability(self, capability_name: str):
        """
        Pause capability processing, preventing future requests from being executed

        :param capability_name: Name of capability to pause
        """
        capability = self.capability_info.lookup_capability(capability_name)
        capability.paused = True
        self.capability_info.save_entity(capability)

    def unpause_capability(self, capability_name: str):
        """
        Unpause capability processing, allowing future requests to be executed normally

        :param capability_name: Name of capability to unpause
        """
        capability = self.capability_info.lookup_capability(capability_name)
        capability.paused = False
        self.capability_info.save_entity(capability)

        # Restart queue manager
        self.execution_manager.queue_manager.startup()

    @on_message(type="execution-complete")
    def complete_request(self, **message: Dict):
        logger.info(f"RECEIVED EXECUTION-COMPLETE: {message}")

        execution = message["subject"]
        capability_request = self.capability_info.lookup_capability_request(execution["capability_request_id"])
        capability_request.state = CapabilityRequestState.Complete.name
        self.capability_info.save_entity(capability_request)

        capability_complete_msg = CapabilityMessageArchitect(request=capability_request).compose_message(
            "capability_complete"
        )
        self.messenger.send_message(**capability_complete_msg)

    @on_message(type="execution-failed")
    def fail_request(self, **message: Dict):
        logger.info(f"RECEIVED EXECUTION-FAILED: {message}")

        execution = message["subject"]
        capability_request = self.capability_info.lookup_capability_request(execution["capability_request_id"])
        capability_request.state = CapabilityRequestState.Failed.name
        self.capability_info.save_entity(capability_request)

        capability_failed_msg = CapabilityMessageArchitect(request=capability_request).compose_message(
            "capability_failed"
        )
        self.messenger.send_message(**capability_failed_msg)

    @on_message(service="workflow", type="delivery")
    def on_delivery(self, **message: Dict):
        """
        Catch a delivery notification and update the locations in the
        referenced execution based on it

        :param message: a delivery-type notification with a workflow request
        subject and a delivery field
        """
        logger.info("RECEIVED DELIVERY NOTIFICATION: %s", message)

        # retrieve the delivery structure in the message
        delivery = message["delivery"]

        # we caught a delivery event, find the associated capability execution
        execution = self.capability_info.lookup_execution_by_workflow_request_id(
            message["subject"]["workflow_request_id"]
        )

        # update some fields on the execution, if possible
        # Extra / at the end added to make nginx happy  -nhertz
        execution.delivery_url = delivery["url"] + "/" if "url" in delivery else None
        execution.delivery_path = delivery["delivered_to"] if "delivered_to" in delivery else None

    @on_message(type="capability-submitted")
    def notify_submitted(self, **message: Dict):
        logger.info(f"RECEIVED CAPABILITY-SUBMITTED: {message}")

        subject = message["subject"]
        execution = self.capability_info.lookup_capability_execution(subject["id"])
        self.notify.notify_submitted(execution.capability_request)
        self.execution_manager.run_capability(execution)

    @on_message(type="capability-complete")
    def notify_complete(self, **message: Dict):
        logger.info(f"RECEIVED CAPABILITY-COMPLETE: {message}")

        subject = message["subject"]
        request = self.capability_info.lookup_capability_request(subject["id"])
        self.notify.notify_complete(request)

    @on_message(type="capability-failed")
    def notify_failed(self, **message: Dict):
        logger.info(f"RECEIVED CAPABILITY-FAILED: {message}")

        subject = message["subject"]
        request = self.capability_info.lookup_capability_request(subject["id"])
        self.notify.notify_failed(request)

    @on_message(type="qa_ready")
    def notify_qa_ready(self, **message: Dict):
        logger.info(f"RECEIVED CAPABILITY MESSAGE: {message}")

        subject = message["subject"]
        execution = self.capability_info.lookup_execution_by_workflow_request_id(subject["workflow_request_id"])
        self.notify.notify_qa_ready(execution.capability_request)

    @on_message(service="workflow", type="ingestion-complete")
    def on_ingestion_complete(self, **message: Dict[str, str]):
        """
        Catch an ingestion-complete message sent from ingest_envoy and update the ingested flag
        of the corresponding capability request

        :param message: Ingestion-complete message
        """
        logger.info(f"RECEIVED INGESTION COMPLETE MESSAGE: {message}")

        execution = self.capability_info.lookup_execution_by_workflow_request_id(
            int(message["subject"]["execution_wf_id"])
        )
        request = self.capability_info.lookup_capability_request(execution.capability_request.id)
        request.ingested = True
        self.capability_info.save_entity(request)

    @on_message(type="carta-ready")
    def on_carta_ready(self, **message: Dict[str, str]):
        """
        Catch the RH-flavored event and save it to the capability request version metadata

        :param message: Ingestion-complete message
        :return:
        """
        logger.info(f"RECEIVED CARTA READY MESSAGE: {message}")

        subject = message["subject"]
        parent_wf_request_id = int(subject["argument"]["parent_wf_request_id"])

        execution = self.capability_info.lookup_execution_by_workflow_request_id(parent_wf_request_id)
        request_version = execution.version
        request_version.workflow_metadata = {"carta_url": message["carta_url"]}

        self.capability_info.save_entity(request_version)


class CapabilityLauncher:
    """
    The CapabilityLauncher provides a way for REST clients to access the server-singleton
    CapabilityService for basically the one thing they need it for, to launch capabilities.

    It's important that we share the message router here with the global one, as otherwise
    messages get round-robined to requests-based CapabilityLaunchers rather than the server-
    singleton CapabilityService which actually knows how to process these events.
    """

    def __init__(self, capability_info: CapabilityInfoIF, message_router: Router):
        # since this is used by the REST API to implement its methods,
        # it's fairly important that it not register itself for on_message replies.
        logger.info("Instantiating capability launcher")
        self.message_router = message_router
        self.capability_info = capability_info

    def launch_capability(self, request: CapabilityRequest) -> CapabilityExecutionIF:
        """
        Use the interface of the execution manager to run a capability
        TODO: Executions should be made from capability versions, not requests

        :param request: Capability request
        :return: Execution of the request
        """
        # first we must create the execution
        execution = self.capability_info.create_execution(request)

        # ensure it's saved
        transaction.commit()

        # now pass it back to the CapabilityService to do the next thing
        capability_submitted_msg = CapabilityMessageArchitect(execution=execution).compose_message(
            "capability_submitted"
        )
        self.message_router.send_message(**capability_submitted_msg)

        return execution
