import logging

from workspaces.capability.services.interfaces import (
    ExecutionManagerIF,
    QueueReporterIF,
)

logger = logging.getLogger(__name__)


class CapabilityQueueManager:
    """
    This is a prototype of a new kind of capability queue manager that does not hold onto any
    internal state itself. All of its calculations are performed against a certain database
    query, which is re-run whenever something "interesting" happens. The API of the class
    is basically to provide a way to send signals when "interesting" things happen, so it
    can react by checking to see if we're ready to execute something or not.
    """

    def __init__(self, queue_reporter: QueueReporterIF, executor: ExecutionManagerIF):
        self.queue_reporter = queue_reporter
        self.executor = executor

    def execution_requested(self):
        self.launch_if_availability()

    def execution_complete(self):
        self.launch_if_availability()

    def launch_if_availability(self):
        logger.info("Obtaining queue report")
        for queue in self.queue_reporter.report():
            logger.info(
                "Considering capability queue %s with %s allowed, %s executing and %s waiting",
                queue.capability,
                queue.concurrency_allowed,
                queue.executing,
                queue.waiting,
            )
            # how many jobs do we start? if we have an arbitrarily large number, we can have
            # as many as there are allowed by the concurrency limit. But some are probably
            # already executing, so we deduct the executing number.
            #
            # What if we don't have enough jobs to fulfill the demand? That's OK, we won't
            # launch more jobs than we have available.
            #
            # Thus, the minimum of what is still allowed and what is available
            jobs_to_run = min(queue.concurrency_allowed - queue.executing, queue.waiting)
            if jobs_to_run > 0:
                for execution in self.queue_reporter.waiting_executions(queue.capability, jobs_to_run):
                    logger.info("Waking up execution %s on step %s", execution.id, execution.current_step_idx)
                    self.executor.execute(execution)

    def reconfigure(self):
        self.startup()

    def startup(self):
        self.launch_if_availability()
